/* LICENSE BLOCK
Part of mde: a Modular D game-oriented Engine
Copyright © 2007-2008 Diggory Hardy

This program is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation, either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/** The content system − string-based editable content.
 */
module mde.content.AStringContent;
public import mde.content.Content;

//FIXME: efficient conversions? Need to dup result when formatting a string anyway?
import Int = tango.text.convert.Integer;
import Float = tango.text.convert.Float;
import derelict.sdl.keysym;

import tango.util.log.Log : Log, Logger;
private Logger logger;
static this () {
    logger = Log.getLogger ("mde.content.AStringContent");
}

/** Base class for content containing a simple value editable as text.
 *
 * Derived classes should implement endEdit to convert sv and assign its value
 * to self, then call endEvent.
 *
 * On assignation by this or opAssign the value should be converted to a string
 * and assigned to sv, and pos should be clamped to [0,sv.length] (i.e. enforce
 * pos <= sv.length). */
abstract class AStringContent : Content
{
    protected this (char[] symbol) {
	super (symbol);
    }
    
    /// Get the text.
    override char[] toString (uint i) {
        return i == 0 ? sv
             : i == 1 ? name_
             : i == 2 ? desc_
             : null;
    }
    
    /** Acts on a keystroke and returns the new value.
     *
     * Supports one-line editing: left/right, home/end, backspace/delete. */
    char[] keyStroke (ushort sym, char[] i) {
	debug assert (i.length, "StringContent.keyStroke: no value (??)");	// impossible?
	char k = *i;
	if (k >= 0x20) {
	    if (k == 0x7f) {		// delete
		size_t p = pos;
		if (p < sv.length) ++p;
		while (p < sv.length && (sv[p] & 0x80) && !(sv[p] & 0x40))
		    ++p;
		sv = sv[0..pos] ~ sv[p..$];
	    } else {			// insert character
		char[] tail = sv[pos..$];
		sv.length = sv.length + i.length;
		size_t npos = pos+i.length;
		if (tail) sv[npos..$] = tail.dup;	// cannot assign with overlapping ranges
		    sv[pos..npos] = i;
		pos = npos;
	    }
	} else {			// use sym; many keys output 0
            if (sym >= SDLK_RSHIFT &&  (sym <= SDLK_COMPOSE || sym == SDLK_MENU)) {
            }	// all modifier keys and contect menu key; should be ignored
	    else if (sym == SDLK_BACKSPACE) {	// backspace; k == 0x8
		char[] tail = sv[pos..$];
		if (pos) --pos;
		while (pos && (sv[pos] & 0x80) && !(sv[pos] & 0x40))
		    --pos;
		sv = sv[0..pos] ~ tail;
	    } else if (sym == SDLK_LEFT) {
		if (pos) --pos;
		while (pos && (sv[pos] & 0x80) && !(sv[pos] & 0x40))
		    --pos;
	    } else if (sym == SDLK_RIGHT) {
		if (pos < sv.length) ++pos;
		while (pos < sv.length && (sv[pos] & 0x80) && !(sv[pos] & 0x40))
		    ++pos;
	    } else if (sym == SDLK_HOME || sym == SDLK_UP) {
		pos = 0;
	    } else if (sym == SDLK_END || sym == SDLK_DOWN) {
		pos = sv.length;
	    } else
		debug logger.trace ("Unhandled symbol: {}", sym);
	}
	return sv;
    }
    
    /** Get the character the edit cursor is in front of.
     *
     * NOTE: unused. */
    size_t editIndex () {
	size_t i = 0;
	for (size_t p = 0; p < pos; ++p)
	    if (!(sv[p] & 0x80) || sv[p] & 0x40)
		++i;
	return i;
    }
    /** Set the edit position in front of i'th character.
     *
     * Assumes at least i characters are present; if not this will work but not be optimal. */
    void editIndex (size_t i) {
        pos = 0;
        for (; i > 0; --i) {	// NOTE: could be slightly optimised
            if (pos < sv.length) ++pos;
            while (pos < sv.length && (sv[pos] & 0x80) && !(sv[pos] & 0x40))
                ++pos;
        }
    }
    
    /** Call after editing a string; return new string (may be changed/reverted). */
    char[] endEdit ();
    
protected:
    char[] sv;		// string of value; updated on assignment for displaying and editing
    size_t pos;		// editing position; used by keyStroke
}

class BoolContent : AStringContent
{
    /** Create a content with _symbol name symbol. */
    this (char[] symbol) {
        auto valp = symbol in changed.boolData;
        if (valp)
            v = *valp;
        sv = v ? "true" : "false";
	super (symbol);
    }
    
    // Assign without adding change to save changeset
    void assignNoCng (bool val) {
	v = val;
	sv = v ? "true" : "false";
	if (pos > sv.length) pos = sv.length;
        endEvent;
    }
    void opAssign (bool val) {
	assignNoCng (val);
        endCng;
    }
    bool opCall () {
        return v;
    }
    alias opCall opCast;
    
    override char[] endEdit () {
	v = sv && (sv[0] == 't' || sv[0] == 'T' || sv[0] == '1');
        sv = v ? "true" : "false";
        endEvent;
        endCng;
	return sv;
    }
    
    // Add change to changeset
    void endCng () {
        changed.boolData[symbol] = v;
    }
    
protected:
    bool v;
}

/** Text content. */
class StringContent : AStringContent
{
    this (char[] symbol) {
        auto valp = symbol in changed.charAData;
        if (valp)
            v = *valp;
        super (symbol);
    }
    
    void assignNoCng (char[] val) {
	v = val;
	if (pos > sv.length) pos = sv.length;
        endEvent;
    }
    void opAssign (char[] val) {
	assignNoCng (val);
        endCng;
    }
    char[] opCall () {
        return v;
    }
    alias opCall opCast;
    
    override char[] endEdit () {
	endEvent;
        endCng;
	return sv;
    }
    
    void endCng () {
        changed.charAData[symbol] = v;
    }
    
protected:
    alias sv v;		// don't need separate v and sv in this case
}

/** Integer content. */
class IntContent : AStringContent
{
    /** Create a content with _symbol name symbol. */
    this (char[] symbol) {
        auto valp = symbol in changed.intData;
        if (valp)
            v = *valp;
        sv = Int.toString (v);
	super (symbol);
    }
    
    void assignNoCng (int val) {
	v = val;
	sv = Int.toString (v);
	if (pos > sv.length) pos = sv.length;
        endEvent;
    }
    void opAssign (int val) {
	assignNoCng (val);
	endCng;
    }
    int opCall () {
        return v;
    }
    alias opCall opCast;
    
    override char[] endEdit () {
	try {
            v = Int.toInt (sv);
        } catch (Exception e) {
            logger.warn (e.msg);
        }
        sv = Int.toString (v);
        endEvent;
        endCng;
	return sv;
    }
    
    void endCng () {
        changed.intData[symbol] = v;
    }
    
protected:
    int v;
}

/** Double content. */
class DoubleContent : AStringContent
{
    /** Create a content with _symbol name symbol. */
    this (char[] symbol) {
        auto valp = symbol in changed.doubleData;
        if (valp)
            v = *valp;
        sv = Float.toString (v, 8, 4);
	super (symbol);
    }
    
    void assignNoCng (double val) {
	v = val;
	sv = Float.toString (v, 8, 4);
	if (pos > sv.length) pos = sv.length;
        endEvent;
    }
    void opAssign (double val) {
	assignNoCng (val);
        endCng;
    }
    double opCall () {
        return v;
    }
    alias opCall opCast;
    
    override char[] endEdit () {
        try {
            v = Float.toFloat (sv);
        } catch (Exception e) {
            logger.warn (e.msg);
        }
        sv = Float.toString (v, 8, 4);
        endEvent;
        endCng;
	return sv;
    }
    
    void endCng () {
        changed.doubleData[symbol] = v;
    }
    
protected:
    double v;
}

/** A content representing an enumeration. */
class EnumContent : AStringContent, IContentList
{
    /** CTOR.
    *
    * Params:
    *	enumSymbols = Symbol names for each
    *	val = which value is active; must be in [0,length-1]
    */
    this (char[] symbol, char[][] enumSymbols) {
        super (symbol);
        this.enumSymbols = enumSymbols;
        enums.length = enumSymbols.length;
        char[] symPeriod = symbol~'.';
        foreach (i, ref e; enums) {
            e = new EnumValueContent (this, i, symPeriod~enumSymbols[i]);
        }
        enums[v].assignFromParent (true);
        sv = enums[v].name_;
        // Re-set the value if a saved value is found:
        auto valp = symbol in changed.enumValData;
        if (valp)
            assignNoCng = *valp;
    }
    
    void opAssign (size_t val) {
        assignNoCng (val);
        endCng;
    }
    // Assign by enum symbol name (for ContentLoader)
    void assignNoCng (char[] enumSym) {
        foreach (i,e; enumSymbols) {
            if (e == enumSym) {
                assignNoCng (i);
                return;
            }
        }
        logger.warn ("EnumContent {} assigned invalid enumeration: {}; valid: {}", symbol, enumSym, enumSymbols);
    }
    size_t opCall () {
        //debug logger.trace ("EnumContent {} returning value: {} ({})",symbol, enumSymbols[v], v);
        return v;
    }
    alias opCall opCast;
    void assignNoCng (size_t val) {
        if (val >= enums.length) {
	    logger.error ("EnumContent "~name_~" assigned invalid value; keeping value: "~sv);
	    return;
	}
        enums[v]  .assignFromParent (false);
        enums[val].assignFromParent (true);
	v = val;
        sv = enums[v].name_;
        if (pos > sv.length) pos = sv.length;
        endEvent;
    }
    
    override char[] endEdit () {
	foreach (i,e; enums)
	    if (sv == e.name_) {
		assignNoCng (i);
		goto break1;
	    }
	
        sv = enums[v].name_;	// sv was edited; revert
        logger.error ("EnumContent "~name_~" assigned invalid value; keeping value: "~sv);
	if (pos > sv.length) pos = sv.length;
	
	break1:
	endCng;
	return sv;
    }
    
    void endCng () {
        changed.enumValData[symbol] = enumSymbols[v];
    }

    override Content[] list () {
        return enums;
    }
    
    // Interface functions that don't make sense for an emuneration:
    override void append (Content) {}
    
protected:
    // Called by child; change this if true, assert current value if false
    void childAssign (size_t val) {
        debug assert (val < enums.length, "cA out of bounds");
        if (enums[val].v)
            this = val;
        else
            enums[val].assignFromParent (v == val);
    }
    
    size_t v;			// value (i.e. enums[v] is value)
    EnumValueContent[] enums;
    char[][] enumSymbols;	// saved for getStructOf
    
    /** Special version of BoolContent for each enumeration to update the
     * parent Enum.
     * 
     * Also should not save its value, since the parent stores the value. */
    private class EnumValueContent : BoolContent {
        /** New enumeration of parent with index num. */
        this (EnumContent parent, size_t num, char[] symbol) {
            this.parent = parent;
            i = num;
            super (symbol);
        }
        
        override void assignNoCng (bool val) {
            v = val;
            parent.childAssign (i);
        }
        override void opAssign (bool val) {
            assignNoCng (val);
        }
        void assignFromParent (bool val) {	// don't call back to parent
            super.assignNoCng (val);
        }
        
        override char[] endEdit () {
            v = sv && (sv[0] == 't' || sv[0] == 'T' || sv[0] == '1');
            parent.childAssign (i);
            return sv;
        }
    
    protected:
        EnumContent parent;
        size_t i;
    }
}
